@@ -118,9 +118,10 @@ def create_metadata_folder(
     # Connect to database and get the table metadata
     create_json_for_tables(
         tables=table_list,
-        database=schema,
+        schema=schema,
         location=destination,
         include_op_column=True,
+        include_timestamp_column=True,
         include_derived_columns=False,
         include_objects=False,
         connection=connection,
@@ -173,9 +174,10 @@ def create_json_for_database(
 
 def create_json_for_tables(
     tables: list,
-    database: str,
+    schema: str,
     location="./meta_data",
     include_op_column=True,
+    include_timestamp_column=True,
     include_derived_columns=False,
     include_objects=False,
     connection=None,
@@ -189,8 +191,8 @@ def create_json_for_tables(
     tables (list):
         A list of table names from the source database.
 
-    database (str):
-        The name of the database.
+    schema (str):
+        The name of the Oracle schema to get metadata from.
 
     location (string):
         Folder where you want to save the metadata.
@@ -199,6 +201,10 @@ def create_json_for_tables(
         If True, adds a column called 'Op' for operation type - I, U or D
         This data is automatically added by AWS DMS ongoing replication tasks.
 
+    include_timestamp_column (string):
+        If True, adds a column called 'extraction_timestamp'.
+        This data is added by DMS if the source extra connection attribute is set.
+
     include_derived_columns (boolean):
         If True, adds 5 datetime columns: mojap_extraction_datetime,
         mojap_start_datetime, mojap_end_datetime, mojap_latest_record, mojap_image_tag.
@@ -213,29 +219,38 @@ def create_json_for_tables(
     """
     make_folder_if_new(location)
     cursor = connection.cursor()
+
+    # Get all the non-nullable columns
+    not_nullable = find_not_nullable(cursor, schema)
+
     problems = {}
     for table in tables:
         try:
             # Get a row to see the columns and check the table has data
             # 'WHERE ROWNUM <= 1' is Oracle for 'LIMIT 1'
             # fetchone() to see the first row of the query executed
-            cursor.execute(f"SELECT * FROM {database}.{table} WHERE ROWNUM <= 1")
+            cursor.execute(f"SELECT * FROM {schema}.{table} WHERE ROWNUM <= 1")
             cursor.fetchone()
             # For Delius purposes we only want tables with at least 1 row
             # This might not be the case for all metadata
             if cursor.rowcount > 0:
                 metadata = get_table_meta(
-                    cursor, table, include_op_column, include_derived_columns
+                    cursor,
+                    table,
+                    not_nullable.get(table.lower(), []),
+                    include_op_column,
+                    include_timestamp_column,
+                    include_derived_columns,
                 )
                 write_json(metadata, Path(location) / f"{table.lower()}.json")
 
             else:
-                print(f"No rows in {table} from {database}")
+                print(f"No rows in {table} from {schema}")
 
         except Exception as e:
             # Likely errors are that the table has been deleted, marked
             # for deletion, or relies on an external reference you can't access
-            print(f"Problem reading {table} in {database}")
+            print(f"Problem reading {table} in {schema}")
             problems[table] = e.args[0].message  # this attribute may be Oracle-only
             continue
 
@@ -252,7 +267,9 @@ def create_json_for_tables(
 def get_table_meta(
     cursor,
     table: str,
+    not_nullable: list,
     include_op_column: bool = True,
+    include_timestamp_column: bool = True,
     include_derived_columns: bool = False,
     include_objects: bool = False,
 ) -> list:
@@ -269,10 +286,17 @@ def get_table_meta(
     table (str):
         Name of the table.
 
+    not_nullable (list):
+        List of column names in this table that aren't nullable.
+
     include_op_column (boolean):
         If True, adds a column called 'Op' for operation type - I, U or D
         This data is automatically added by AWS DMS ongoing replication tasks.
 
+    include_timestamp_column (boolean):
+        If True, adds a column called 'EXTRACTION_TIMESTAMP'
+        This data is added by AWS DMS, when configured as an extra connection attribute.
+
     include_derived_columns (boolean):
         If True, adds 5 datetime columns: mojap_extraction_datetime,
         mojap_start_datetime, mojap_end_datetime, mojap_latest_record, mojap_image_tag.
@@ -288,13 +312,13 @@ def get_table_meta(
     """
     # This lookup is specific to Oracle data types
     # Likely to need separate dictionaries for other data sources
+    # Numbers aren't covered here - they're handled further down as they're more complex
     type_lookup = {
         "DB_TYPE_DATE": "datetime",
         "DB_TYPE_TIMESTAMP": "datetime",
         "DB_TYPE_TIMESTAMP_TZ": "datetime",
         "DB_TYPE_CHAR": "character",
         "DB_TYPE_CLOB": "character",
-        "DB_TYPE_BLOB": "character",
         "DB_TYPE_VARCHAR": "character",
         "DB_TYPE_LONG": "character",
         "DB_TYPE_RAW": "character",
@@ -313,27 +337,29 @@ def get_table_meta(
             }
         )
 
+    if include_timestamp_column:
+        columns.append(
+            {
+                "name": "EXTRACTION_TIMESTAMP",
+                "type": "character",
+                "description": "DMS extraction timestamp",
+                "nullable": False,
+            }
+        )
+
     # Data types to skip. This list is specific to working with
     # Amazon Database Migration Service and an Oracle database.
     # These are the Oracle datatypes that DMS can't copy
-    if include_objects:
-        skip = [
-            "DB_TYPE_ROWID",
-            "DB_TYPE_BFILE",
-            "REF",
-            "ANYDATA",
-        ]
-    else:
-        skip = [
-            "DB_TYPE_ROWID",
-            "DB_TYPE_BFILE",
-            "REF",
-            "ANYDATA",
-            "DB_TYPE_OBJECT",
-        ]
+    # Plus blob columns, which we're intentionally excluding
+    skip = ["REF", "ANYDATA", "DB_TYPE_ROWID", "DB_TYPE_BFILE", "DB_TYPE_BLOB"]
+
+    if not include_objects:
+        skip.append("DB_TYPE_OBJECT")
+
     # Main column info - cursor.description has 7 set columns:
     # name, type, display_size, internal_size, precision, scale, null_ok
     # Usually type is just a lookup, but decimal types need scale and precision too
+    # null_ok isn't accurate in Delius, so nullable uses an external dict
     for col in cursor.description:
         if col[1].name not in skip:
             columns.append(
@@ -343,7 +369,7 @@ def get_table_meta(
                     if col[1].name == "DB_TYPE_NUMBER"
                     else type_lookup[col[1].name],
                     "description": "",
-                    "nullable": bool(col[6]),
+                    "nullable": bool(col[0].lower() not in not_nullable),
                 }
             )
 
@@ -490,3 +516,41 @@ def get_subpartitions(table, partition, cursor):
     result = cursor.fetchall()
     subpartitions = [item[0] for item in result]
     return subpartitions
+
+
+def find_not_nullable(cursor, schema):
+    """Returns a dictionary of tables, each with a list of their non-nullable columns.
+
+    Can't use table descriptions for this as the null_ok field is inaccurate in Delius.
+
+    Parameters
+    ----------
+    cursor:
+        A database Cursor object ready to execute queries with.
+
+    schema (str):
+        Name of the database schema you want to get data from.
+        This is schema in the Oracle sense of a section of the database.
+    """
+    # SQL to get all primary key (P) and check (C) constraints from the selected schema
+    cursor.execute(
+        f"""
+        SELECT
+            cons.table_name,
+            cols.column_name,
+            cons.constraint_type,
+            cons.search_condition
+        FROM all_constraints cons, all_cons_columns cols
+        WHERE cons.constraint_name = cols.constraint_name
+        AND cons.owner = '{schema}'
+        AND (cons.constraint_type = 'C' OR cons.constraint_type = 'P')
+        """
+    )
+    constraints = cursor.fetchall()
+    not_nullable = {}
+    for con in constraints:
+        # For 'C' constraints, only want the 'IS NOT NULL' ones
+        # Can't put this check in the SQL query because can't use LIKE on LONG datatype
+        if con[2] == "P" or con[3].endswith("IS NOT NULL"):
+            not_nullable.setdefault(con[0].lower(), []).append(con[1].lower())
+    return not_nullable
